/*
   Copyright (C) 2003 Paul Brossier

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/** \bug works only on 44100Hz signal */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <unistd.h>

#include <math.h> /**< \bug required for finite check in averaging */

#include "aubio.h"	

#include "utils.h"

/* our processing function */
int aubio_process(float **input, float **output, int nframes);

/* default settings */
/* possible onset types: energy,specdiff,hfc,complexdomain,phase */
//aubio_onsetdetection_type type_onset = hfc;
aubio_onsetdetection_type type_onset  = hfc;
aubio_onsetdetection_type type_onset2 = complexdomain;
int averaging                = 0;
int verbose                  = 0;
int usejack                  = 0;
int buffer_size              = 1024;
int overlap_size             = 512;
int channels                 = 1;
int samplerate               = 44100;
int isonset                  = 0;
int wasnote                  = 0;
	uint_t note_curpos = 0;
smpl_t threshold             = 0.1;
smpl_t threshold2            = -80.;
uint_t note_winlength        = 10;
const char * output_filename = NULL;
const char * input_filename  = NULL;
const char * onset_filename  = "/usr/share/sounds/aubio/woodblock.aiff";

/* global objects */
aubio_pvoc_t * pv            = NULL;
aubio_pvoc_t * pv_pitch      = NULL;
aubio_filter_t * filter      = NULL;
fvec_t * ibuf                = NULL;
fvec_t * obuf                = NULL;
fvec_t * woodblock           = NULL;
cvec_t * fftgrain            = NULL;
/*
   fvec_t *onset              = NULL;
   aubio_onsetdetection_t *o  = NULL;
   */
aubio_onsetdetection_t *o1;
aubio_onsetdetection_t *o2;
fvec_t *onset1;
fvec_t *onset2;
aubio_pickpeak_t * parms   = NULL;
cvec_t * fftgrain_pitch    = NULL;
cvec_t * fftgrain_memory   = NULL;
aubio_pitchdetection_t * p = NULL;
smpl_t pitch               = 0;
smpl_t note_pitch          = 0;
int wassilence             = 0;
int offset                 = 0;
int issilence              = 0;
smpl_t weight              = 1.;

fvec_t * note_buffer       = NULL;

int aubio_fftaveraging(cvec_t * input_new, cvec_t * output_accum, smpl_t weight);
int aubio_fftaveraging(cvec_t * input_new, cvec_t * output_accum, smpl_t weight) {
	uint_t i, j;
	for ( i = 0; i < output_accum->channels ; i++ ) {
		for ( j = 0; j < output_accum->length ; j++ ) {
			if (finite(output_accum->norm[i][j]))
				output_accum->norm[i][j] =
					(1.-weight) * output_accum->norm[i][j]
					+ weight    * input_new->norm[i][j] ;
			else 
				output_accum->norm[i][j] = input_new->norm[i][j];
		}}
	return 0;
}

int aubio_getnote(fvec_t * note_buffer, uint_t note_curpos);
int aubio_getnote(fvec_t * note_buffer, uint_t note_curpos){
	return vec_median(note_buffer);
}



int aubio_process(float **input, float **output, int nframes) {
	/** \bug the block processing is redundant in the offline mode */
	uint_t i=0; 			/*channels*/
	uint_t j=0; 			/*frames*/   
	uint_t pos = 0; /*frames%dspblocksize*/

	for (j=0;j<nframes;j++) {

		for (i=0;i<channels;i++) { 
			/* write input to datanew */
			fvec_write_sample(ibuf, input[i][j], i, pos);
			/* put synthnew in output */
			output[i][j] = fvec_read_sample(obuf, i, pos);
		}
		/*time for fft*/
		if (pos == overlap_size-1) {

			/* test for silence */
			issilence = aubio_silence_detection(ibuf, threshold2);
			/* currently silence */
			if (issilence && wassilence == 1) {
				offset = 0;
				/* new offset */
			} else if (issilence && wassilence == 0) {
				wassilence = 1; offset = 1;
				/* end of silence */
			} else {
				wassilence = 0; offset = 0;
			}

			/* onset phase voc */
			aubio_pvoc_do (pv,ibuf, fftgrain);
			/* onset detection function */
			//aubio_onsetdetection(o,fftgrain, onset);
			aubio_onsetdetection(o1,fftgrain, onset1);
			aubio_onsetdetection(o2,fftgrain, onset2);
			onset1->data[0][0] *= onset2->data[0][0];
			/* onset peack picking */
			//isonset = aubio_peakpick_pimrt(onset,parms);
			isonset = aubio_peakpick_pimrt(onset1,parms);

			/* pitch pre processing */
			aubio_filter_do(filter,ibuf);
			aubio_filter_do(filter,ibuf);
			/* pitch phase voc */
			aubio_pvoc_do (pv_pitch,ibuf, fftgrain_pitch);
			/* pitch fft averaging (optional, -a flag) */
			if (averaging) {
				if (isonset) weight = 1.;
				else         weight += 1.;
				aubio_fftaveraging(
						fftgrain_pitch, 
						fftgrain_memory, 
						1./weight * (!issilence));
				pitch = aubio_pitchdetect(p,fftgrain_memory);
			} else {
				pitch = aubio_pitchdetect(p,fftgrain_pitch);
			}
			/* note segmentation */

			wasnote = 0;
			if (!issilence) {
				//outmsg("%d \t", note_curpos);
				if (note_curpos < note_winlength) {
					note_buffer->data[0][note_curpos] = pitch;
					if (isonset &&  note_curpos > 4) {
						/* if onset before note_winlength, output note */
						//if (note_curpos < note_winlength)
						wasnote = 1;
						note_pitch = aubio_getnote(note_buffer,note_curpos);
						/* reset the counter (onset occured) */
						note_curpos = 0;
						/* empty the buffer */
						for (pos=0; pos < note_winlength; pos++)
							note_buffer->data[0][pos] = 0;
					} else {
						note_curpos += 1;
					}
				} else if (!wasnote && note_curpos == note_winlength) {
					wasnote = 1;
					note_pitch = aubio_getnote(note_buffer,note_curpos);
					note_curpos = 0;
					for (pos=0; pos < note_winlength; pos++)
						note_buffer->data[0][pos] = 0;
				}
			}

			if (isonset)
				note_curpos = 0;

			/* notes audio output */
			if (usejack || (output_filename != NULL)){
				if (isonset) {
					for (pos = 0; pos < overlap_size; pos++)
						obuf->data[0][pos] = woodblock->data[0][pos];
				} else {
					for (pos = 0; pos < overlap_size; pos++)
						obuf->data[0][pos] = 0.;
				}
			}
			//aubio_pvoc_rdo(pv,fftgrain, obuf);
			pos = -1; /* so it will be zero next j loop */
		}
		pos++;
	}
	return 1;
}

int main (int argc, char **argv) {  
	int frames;
	//int pos;
	aubio_file_t * file = NULL;
	aubio_file_t * fileout = NULL;
	aubio_file_t * onsetfile = new_file_ro(onset_filename);
	/* command line argument parsing */
	parse_args(argc, argv);
	debug("type_onset: %d\n", (int)type_onset);
	debug("type_onset2: %d\n", (int)type_onset2);
	debug("threshold : %f\n", threshold);

	if(!usejack) 
	{
		file = new_file_ro (input_filename);
		file_info(file);
		/** \bug: should set samplerate here too */
		channels = file_channels(file);
		if (output_filename != NULL)
			fileout = new_file_wo(file, output_filename);
	}

	ibuf = new_fvec(overlap_size, channels);
	obuf = new_fvec(overlap_size, channels);
	note_buffer = new_fvec(note_winlength, channels);
	/** \bug order counts ?! */
	fftgrain        = new_cvec(buffer_size, channels);
	fftgrain_pitch  = new_cvec(buffer_size*4, channels);
	fftgrain_memory = new_cvec(buffer_size*4, channels);
	pv       = new_aubio_pvoc(buffer_size, overlap_size, channels);
	pv_pitch = new_aubio_pvoc(buffer_size*4, overlap_size, channels);

	if (usejack || output_filename != NULL ) {
		woodblock = new_fvec(buffer_size,1);
		file_read(onsetfile, overlap_size, woodblock);
	}

	filter = new_aubio_adsgn_filter((smpl_t)samplerate);

	p      = new_aubio_pitchdetection(buffer_size,channels);
	parms  = new_aubio_peakpicker(threshold);

	/* onset init */
	//o      = new_aubio_onsetdetection(type_onset,buffer_size,channels);
	//onset  = new_fvec(overlap_size, channels);
	o1 = new_aubio_onsetdetection(type_onset,buffer_size,channels);
	onset1 = new_fvec(1, channels);
	o2 = new_aubio_onsetdetection(type_onset2,buffer_size,channels);
	onset2 = new_fvec(1, channels);

	if(usejack) {
#ifdef JACK_SUPPORT
		aubio_jack_t * jack_setup;
		debug("Jack init ...\n");
		jack_setup = new_aubio_jack(channels, channels, (aubio_process_func_t)aubio_process);

		/* initialise fftgrain */
		aubio_pvoc_do (pv,ibuf, fftgrain);

		debug("Jack activation ...\n");
		aubio_jack_activate(jack_setup);

		debug("Processing (Ctrl+C to quit) ...\n");
		pause();

		aubio_jack_close(jack_setup);
#endif

	} else {
		fvec_t * inputbuf ;
		fvec_t * outputbuf;
		/* file objects */

		inputbuf 	= new_fvec(overlap_size, channels);
		outputbuf	= new_fvec(overlap_size, channels);	
		/* phasevoc init */

		/** \bug fftgrain still needs to be initialised */
		for (frames=0;frames<4;frames++) {
			aubio_pvoc_do (pv,inputbuf, fftgrain);
			aubio_pvoc_do (pv_pitch,inputbuf, fftgrain_pitch);
			aubio_pvoc_do (pv_pitch,inputbuf, fftgrain_memory);
		}

		for (frames=0;frames<1;frames++) {
			file_read(file, overlap_size, inputbuf);
			aubio_pvoc_do (pv,inputbuf, fftgrain);
			aubio_filter_do(filter,inputbuf);
			aubio_pvoc_do (pv_pitch,inputbuf, fftgrain_pitch);
			aubio_pvoc_do (pv_pitch,inputbuf, fftgrain_memory);
		}

		debug("Processing ...\n");
		frames = 0;
		while ((overlap_size == file_read(file, overlap_size, inputbuf)) ) 
		{
			isonset=0;
			/* run process function */
			aubio_process(inputbuf->data, outputbuf->data, overlap_size);
			/* data text output (seconds, time, pitchcand, isonset, issilence */
			/* outmsg("%f %f %d %d\n",
					frames*overlap_size/(smpl_t)samplerate,
					bintomidi(pitch,samplerate,buffer_size*4),
					//pitch,
					isonset,
					wassilence);
					*/
			if (wasnote)
				outmsg("%f %f\n",
						(frames-note_curpos-3)*overlap_size/(smpl_t)samplerate,
						bintomidi(note_pitch,samplerate,buffer_size*4));
			if (output_filename != NULL)
				file_write(fileout,overlap_size,outputbuf);
			frames++;
		}
		debug("Processed %d frames of %d samples.\n", frames, buffer_size);
		if (output_filename != NULL)
			del_file(fileout);
		del_file(file);
	}

	del_aubio_pvoc(pv);
	del_aubio_pvoc(pv_pitch);
	del_fvec(obuf);
	del_fvec(ibuf);
	del_cvec(fftgrain);
	del_cvec(fftgrain_pitch);
	del_cvec(fftgrain_memory);
	//del_fvec(onset);
	del_fvec(onset1);
	del_fvec(note_buffer);
	del_fvec(onset2);

	debug("End of program.\n");
	fflush(stderr);
	return 0;
}
